"""The module contains a set of API for HTTP responses types."""
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Union
import requests

JsonType = Union[Dict[Any, Any], Any]


class ResponseError(Exception):
    """The class represents HTTP api request response error."""

    pass


class Response(ABC):
    """The class represents an abstraction of a response from an API request."""

    @abstractmethod
    def is_ok(self) -> bool:
        """Returns `True` if response is `OK` otherwise `False`."""
        pass

    @abstractmethod
    def code(self) -> int:
        """Returns HTTP response status code."""
        pass

    @abstractmethod
    def as_json(self) -> JsonType:
        """Returns HTTP response data as dictionary type."""
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Returns HTTP response data as plain data type."""
        pass


class HttpResponse(Response):
    """The class represents an HTTP response from HTTP API request."""

    def __init__(self, response: requests.Response) -> None:
        self._response: requests.Response = response

    def is_ok(self) -> bool:
        return self._response.ok

    def code(self) -> int:
        return self._response.status_code

    def as_json(self) -> JsonType:
        return self._response.json()

    def __str__(self) -> str:
        return self._response.text


def safe_response(response: Response, success_codes: Iterable[int] = (200, 201, 204)) -> Response:
    """Specifies safe response from iterable of success HTTP status codes.

    Args:
        response: a specific HTTP response
        success_codes: a list of allowed success codes

    Raises:
        `ResponseError` if HTTP response contains a set of errors

    Returns: a response
    """
    if response.code() not in success_codes:
        raise ResponseError(
            f"HTTP response contains some errors with '{response.code()}' status code! Reason: {response}"
        )
    return response



"""The module contains a set of API for HTTP sessions."""
from abc import abstractmethod
from types import TracebackType
from typing import Any, Dict, Optional, Type
from punish.type import AbstractContextManager
import requests
from requests.auth import HTTPBasicAuth
from urequest.credentials import Credentials  # noqa: I100
from urequest.response import HttpResponse, Response, safe_response
from urequest.url import Address


class Session(AbstractContextManager):
    """The class represents abstract interfaces for an API Session."""

    @abstractmethod
    def get(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``GET`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments

        Returns: response element
        """
        pass

    @abstractmethod
    def options(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``OPTIONS`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments

        Returns: response element
        """
        pass

    @abstractmethod
    def head(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``HEAD`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments

        Returns: response element
        """
        pass

    @abstractmethod
    def post(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        """Performs ``POST`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            plain: requested data as a plain text
            as_dict: requested data as dictionary (json)
            kwargs: other keyword arguments

        Returns: response element
        """
        pass

    @abstractmethod
    def put(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        """Performs ``PUT`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            plain: requested data as a plain text
            as_dict: requested data as dictionary (json)
            kwargs: other keyword arguments

        Returns: response element
        """
        pass

    @abstractmethod
    def patch(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        """Performs ``PATCH`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            plain: requested data as a plain text
            as_dict: requested data as dictionary (json)
            kwargs: other keyword arguments

        Returns: response element
        """
        pass

    @abstractmethod
    def delete(self, url: Address, **kwargs: Any) -> Response:
        """Performs ``DELETE`` HTTP request of a session.

        Args:
            url: url path used to perform a request
            kwargs: keyword arguments

        Returns: response element
        """
        pass


class HttpSession(Session):
    """The class provides interfaces for current API HTTP session."""

    def __init__(self, session: requests.Session = requests.Session()) -> None:
        self._session: requests.Session = session

    def __enter__(self) -> Session:
        return self

    def get(self, url: Address, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.get(str(url), **kwargs)))

    def options(self, url: Address, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.options(str(url), **kwargs)))

    def head(self, url: Address, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.head(str(url), **kwargs)))

    def post(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.post(str(url), data=plain, json=as_dict, **kwargs)))

    def put(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.put(str(url), data=plain, json=as_dict, **kwargs)))

    def patch(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.patch(str(url), data=plain, json=as_dict, **kwargs)))

    def delete(self, url: Address, **kwargs: Any) -> Response:
        return safe_response(HttpResponse(self._session.delete(str(url), **kwargs)))

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._session.close()


class LoggedHttpSession(Session):
    """The class provides logged HTTP session."""

    def __init__(self, credentials: Credentials, session: requests.Session = requests.Session()) -> None:
        session.auth = HTTPBasicAuth(credentials.username, credentials.password)
        self._session: Session = HttpSession(session)

    def __enter__(self) -> Any:
        return self._session.__enter__()

    def get(self, url: Address, **kwargs: Any) -> Response:
        return self._session.get(url, **kwargs)

    def options(self, url: Address, **kwargs: Any) -> Response:
        return self._session.options(url, **kwargs)

    def head(self, url: Address, **kwargs: Any) -> Response:
        return self._session.head(url, **kwargs)

    def post(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        return self._session.post(url, plain, as_dict, **kwargs)

    def put(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        return self._session.post(url, plain, as_dict, **kwargs)

    def patch(self, url: Address, plain: str = None, as_dict: Dict[Any, Any] = None, **kwargs: Any) -> Response:
        return self._session.patch(url, plain, as_dict, **kwargs)

    def delete(self, url: Address, **kwargs: Any) -> Response:
        return self._session.delete(url, **kwargs)

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._session.__exit__(exception_type, exception_value, traceback)



"""Package setup entrypoint."""
from typing import IO, Sequence
from setuptools import find_packages as __find_packages, setup as __compose_package
from urequest import (
    __author__ as __author,
    __doc__ as __short_doc,
    __email__ as __email,
    __license__ as __license,
    __name__ as __name,
    __version__ as __version,
)


def __readme() -> str:
    """Returns project description."""
    with open("README.md") as readme:  # type: IO[str]
        return readme.read()


def __requirements() -> Sequence[str]:
    """Returns requirements sequence."""
    with open("requirements.txt") as requirements:  # type: IO[str]
        return tuple(map(str.strip, requirements.readlines()))


if __name__ == "__main__":
    __compose_package(
        name=__name,
        version=__version,
        author=__author,
        author_email=__email,
        description=__short_doc,
        long_description=__readme(),
        long_description_content_type="text/markdown",
        url="https://github.com/vyahello/urequest",
        packages=__find_packages(exclude=("*.tests", "*.tests.*", "tests.*", "tests")),
        include_package_data=True,
        install_requires=__requirements(),
        classifiers=(
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            f"License :: OSI Approved :: {__license}",
            "Operating System :: OS Independent",
        ),
        python_requires=">=3.6",
    )
